"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
const graphql_1 = require("graphql");
const graphql_tools_1 = require("graphql-tools");
const errors_1 = require("graphql-tools/dist/stitching/errors");
const lodash_1 = require("lodash");
function mergeObjectTypes({ types, newTypes }) {
    const mergeQuery = lodash_1.camelCase(types[0].type.name);
    return new graphql_1.GraphQLObjectType({
        name: types.map(type => type.type.name).filter(name => name)[0],
        description: types.map(type => type.type.description).filter(d => d)[0],
        isTypeOf: types.map(type => type.type.isTypeOf).filter(a => a)[0],
        astNode: types.map(type => type.type.astNode).filter(a => a)[0],
        extensionASTNodes: types
            .map(type => type.type.extensionASTNodes)
            .filter(a => a)[0],
        fields: () => createFieldMapConfig({ types, newTypes, mergeQuery }),
        interfaces: () => mergeInterfaces({ types: types.map(type => type.type), newTypes })
    });
}
function mergeInterfaces({ types, newTypes }) {
    const interfaceMap = {};
    for (const type of types) {
        for (const i of type.getInterfaces()) {
            if (!interfaceMap[i.name]) {
                interfaceMap[i.name] = newTypes[i.name]
                    ? newTypes[i.name]
                    : i;
            }
        }
    }
    return Object.values(interfaceMap);
}
function mergeRootTypes({ types, newTypes }) {
    if (types.length === 0) {
        return undefined;
    }
    else {
        return new graphql_1.GraphQLObjectType({
            name: types.map(type => type.type.name).filter(name => name)[0],
            description: types.map(type => type.type.description).filter(d => d)[0],
            astNode: types.map(type => type.type.astNode).filter(a => a)[0],
            fields: () => createRootFieldMapConfig({ types, newTypes })
        });
    }
}
function createRootFieldMapConfig({ types, newTypes }) {
    const fields = {};
    for (const { type, schema } of types) {
        for (const [key, field] of Object.entries(type.getFields())) {
            if (!fields[key]) {
                fields[key] = [];
            }
            fields[key].push({ schema, field });
        }
    }
    const fieldsConfig = {};
    for (const [key, fieldAndSchema] of Object.entries(fields)) {
        const fieldCandidates = fieldAndSchema.map(f => f.field);
        const schemas = fieldAndSchema.map(f => f.schema);
        const fieldType = getCandidateAttribute(fieldCandidates, "type");
        fieldsConfig[key] = {
            type: newTypes[graphql_1.getNamedType(fieldType).name]
                ? createFieldType(fieldType, newTypes)
                : fieldType,
            args: createArgumentConfig(getCandidateAttribute(Object.values(fieldCandidates), "args"), newTypes),
            resolve: createRootResolver({ schema: schemas[0] }),
            deprecationReason: getCandidateAttribute(Object.values(fieldCandidates), "deprecationReason"),
            description: getCandidateAttribute(Object.values(fieldCandidates), "description"),
            astNode: getCandidateAttribute(Object.values(fieldCandidates), "astNode")
        };
    }
    return fieldsConfig;
}
function getCandidateAttribute(candidates, attribute) {
    const attributes = candidates
        .map(candidate => candidate[attribute])
        .filter(value => value);
    return attributes[0];
}
function createFieldMapConfig({ types, newTypes, mergeQuery }) {
    const fields = {};
    for (const { type, schema, replaceResolvers } of types) {
        for (const [key, field] of Object.entries(type.getFields())) {
            if (!fields[key]) {
                fields[key] = { schema, field, replaceResolvers };
            }
        }
    }
    const fieldsConfig = {};
    for (const [key, { field, schema, replaceResolvers }] of Object.entries(fields)) {
        fieldsConfig[key] = {
            type: newTypes[graphql_1.getNamedType(field.type).name]
                ? createFieldType(field.type, newTypes)
                : field.type,
            args: createArgumentConfig(field.args, newTypes),
            resolve: replaceResolvers
                ? createFieldResolver(schema, mergeQuery)
                : field.resolve,
            deprecationReason: field.deprecationReason,
            description: field.description,
            astNode: field.astNode
        };
    }
    return fieldsConfig;
}
function createRootResolver({ schema }) {
    return (parent, args, context, info) => {
        return graphql_tools_1.delegateToSchema({
            schema,
            operation: info.operation.operation,
            fieldName: info.fieldName,
            args,
            context,
            info
        });
    };
}
function createFieldResolver(schema, mergeQuery) {
    return (parent, args, context, info) => {
        const responseKey = info.fieldNodes[0].alias
            ? info.fieldNodes[0].alias.value
            : info.fieldName;
        const result = parent && parent[responseKey];
        if (result !== undefined) {
            const subErrors = parent[errors_1.ERROR_SYMBOL] &&
                parent[errors_1.ERROR_SYMBOL].filter((error) => error.path && error.path[0] === responseKey).map(({ message, locations, path }) => ({
                    message,
                    locations,
                    path: path && path.slice(1)
                }));
            const fieldError = (subErrors || []).filter((error) => error.path && error.path.length === 0);
            if (fieldError.length > 0) {
                throw fieldError[0].message;
            }
            if (subErrors && subErrors.length > 0) {
                if (Array.isArray(result)) {
                    for (const { message, locations, path } of subErrors) {
                        if (!result[path[0]][errors_1.ERROR_SYMBOL]) {
                            result[path[0]][errors_1.ERROR_SYMBOL] = [];
                        }
                        result[path[0]][errors_1.ERROR_SYMBOL].push({
                            message,
                            locations,
                            path: path.slice(1)
                        });
                    }
                }
                else {
                    result[errors_1.ERROR_SYMBOL] = subErrors;
                }
            }
            return result;
        }
        else {
            if (mergeQuery) {
                info = Object.assign({}, info, { fieldNodes: [
                        {
                            kind: "Field",
                            name: {
                                kind: "Name",
                                value: mergeQuery
                            },
                            selectionSet: {
                                kind: "SelectionSet",
                                selections: info.fieldNodes
                            }
                        }
                    ] });
            }
            const request = graphql_tools_1.delegateToSchema({
                schema,
                operation: "query",
                fieldName: mergeQuery || info.fieldName,
                args: mergeQuery ? { id: parent.id } : args,
                context,
                info
            });
            if (mergeQuery) {
                return request.then(requestResult => requestResult ? requestResult[responseKey] : null);
            }
            else {
                return request;
            }
        }
    };
}
function createFieldType(type, newTypes) {
    if (graphql_1.isNonNullType(type)) {
        return new graphql_1.GraphQLNonNull(createFieldType(type.ofType, newTypes));
    }
    else if (graphql_1.isListType(type)) {
        return new graphql_1.GraphQLList(createFieldType(type.ofType, newTypes));
    }
    else {
        return newTypes[type.name];
    }
}
function createInputFieldType(type, newTypes) {
    if (graphql_1.isNonNullType(type)) {
        return new graphql_1.GraphQLNonNull(createInputFieldType(type.ofType, newTypes));
    }
    else if (graphql_1.isListType(type)) {
        return new graphql_1.GraphQLList(createInputFieldType(type.ofType, newTypes));
    }
    else {
        return newTypes[type.name];
    }
}
function recreateUnionType(type, newTypes) {
    return new graphql_1.GraphQLUnionType({
        name: type.name,
        description: type.description,
        astNode: type.astNode,
        extensionASTNodes: type.extensionASTNodes,
        types: () => type.getTypes().map(t => (newTypes[t.name] || t)),
        resolveType: type.resolveType
    });
}
function recreateScalarType(type) {
    return new graphql_1.GraphQLScalarType({
        name: type.name,
        description: type.description,
        astNode: type.astNode,
        serialize(value) {
            return value;
        },
        parseValue(value) {
            return value;
        }
    });
}
function recreateInputType(type, newTypes) {
    return new graphql_1.GraphQLInputObjectType({
        name: type.name,
        description: type.description,
        astNode: type.astNode,
        fields: () => inputFieldMapToFieldConfigMap(type.getFields(), newTypes)
    });
}
function inputFieldMapToFieldConfigMap(fields, newTypes) {
    const fieldsConfig = {};
    for (const [key, field] of Object.entries(fields)) {
        fieldsConfig[key] = {
            type: newTypes[graphql_1.getNamedType(field.type).name]
                ? createInputFieldType(field.type, newTypes)
                : field.type,
            defaultValue: field.defaultValue,
            description: field.description,
            astNode: field.astNode
        };
    }
    return fieldsConfig;
}
function recreateInterfaceType(type, newTypes) {
    return new graphql_1.GraphQLInterfaceType({
        name: type.name,
        description: type.description,
        astNode: type.astNode,
        extensionASTNodes: type.extensionASTNodes,
        fields: () => recreateInterfaceFieldMap(type.getFields(), newTypes)
    });
}
function recreateInterfaceFieldMap(fields, newTypes) {
    const fieldsConfig = {};
    for (const [key, field] of Object.entries(fields)) {
        fieldsConfig[key] = {
            type: newTypes[graphql_1.getNamedType(field.type).name]
                ? createFieldType(field.type, newTypes)
                : field.type,
            args: createArgumentConfig(field.args, newTypes),
            deprecationReason: field.deprecationReason,
            description: field.description,
            astNode: field.astNode
        };
    }
    return fieldsConfig;
}
function createArgumentConfig(args, newTypes) {
    const argsConfig = {};
    for (const arg of args) {
        argsConfig[arg.name] = {
            type: newTypes[graphql_1.getNamedType(arg.type).name]
                ? createInputFieldType(arg.type, newTypes)
                : arg.type,
            defaultValue: arg.defaultValue,
            description: arg.description,
            astNode: arg.astNode
        };
    }
    return argsConfig;
}
function isTypeToInclude(type) {
    return (type.name !== "Query" &&
        type.name !== "Mutation" &&
        type.name !== "Subscription" &&
        graphql_1.getNamedType(type).name.slice(0, 2) !== "__" &&
        (!graphql_1.isScalarType(type) ||
            (graphql_1.isScalarType(type) && !graphql_1.isSpecifiedScalarType(type))));
}
function mergeRemoteSchemas({ schemas, localSchema }) {
    const newTypes = {};
    const allSchemas = localSchema ? [localSchema, ...schemas] : schemas;
    const queryTypes = allSchemas.map(schema => ({
        schema,
        type: schema.getQueryType()
    }));
    const query = mergeRootTypes({
        types: queryTypes.filter(argument => argument.type),
        newTypes
    });
    const mutationTypes = allSchemas.map(schema => ({
        schema,
        type: schema.getMutationType()
    }));
    const mutation = mergeRootTypes({
        types: mutationTypes.filter(argument => argument.type),
        newTypes
    });
    const typeNameToTypes = {};
    if (localSchema) {
        for (const [key, type] of Object.entries(localSchema.getTypeMap())) {
            if (!typeNameToTypes[key]) {
                typeNameToTypes[key] = [];
            }
            typeNameToTypes[key].push({
                schema: localSchema,
                type,
                replaceResolvers: false
            });
        }
    }
    for (const schema of schemas) {
        for (const [key, type] of Object.entries(schema.getTypeMap())) {
            if (!typeNameToTypes[key]) {
                typeNameToTypes[key] = [];
            }
            typeNameToTypes[key].push({ schema, type, replaceResolvers: true });
        }
    }
    for (const candidates of Object.values(typeNameToTypes)) {
        if (candidates.every(({ type }) => isTypeToInclude(type))) {
            if (candidates.every(({ type }) => graphql_1.isObjectType(type))) {
                const newType = mergeObjectTypes({
                    types: candidates,
                    newTypes
                });
                newTypes[newType.name] = newType;
            }
            else {
                if (candidates.some(({ type }) => graphql_1.isObjectType(type))) {
                    throw new Error(`Can't merge non-Object type ${candidates[0].type.name} with Object type of same name`);
                }
                else {
                    const type = candidates[0].type;
                    if (graphql_1.isUnionType(type)) {
                        newTypes[type.name] = recreateUnionType(type, newTypes);
                    }
                    else if (graphql_1.isInterfaceType(type)) {
                        newTypes[type.name] = recreateInterfaceType(type, newTypes);
                    }
                    else if (graphql_1.isScalarType(type)) {
                        newTypes[type.name] = recreateScalarType(type);
                    }
                    else if (graphql_1.isInputObjectType(type)) {
                        newTypes[type.name] = recreateInputType(type, newTypes);
                    }
                    else {
                        newTypes[type.name] = type;
                    }
                }
            }
        }
    }
    return new graphql_1.GraphQLSchema({
        query,
        mutation,
        types: Object.values(newTypes)
    });
}
exports.mergeRemoteSchemas = mergeRemoteSchemas;
